#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

/* This file contains some utility routines for mean subtraction. */

/* We do the FFT and IFFT using Joerg Arndt's FXT library. */
#include "fxt/src/fft/fft.h"

/* We do audio file I/O using Dan Ellis' snd library. */
extern "C" 
{
#include <snd.h>
#include <sndf.h>
}

#include "meansub-util.h"

/* This function prints an error message and exits. */
void
fatal(char *s, ...) {
  va_list ap;
  va_start(ap, s);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, s, ap);
  fputc('\n', stderr);
  exit(1);
}

/* This function prints a status message. */
void
status(char *s, ...) {
  va_list ap;
  va_start(ap, s);
  vprintf(s, ap);
}

/* This function rounds off a floating-point number */
double round(double x) {
  assert(x >= 0);
  return floor(x+0.5);
}

/* This functions returns 1 if the argument is a power of 2,
   and 0 otherwise. */
int isPowerOf2(unsigned long v) {
  // This code is taken from
  // http://graphics.stanford.edu/~seander/bithacks.html
  return !(v & (v - 1)) && v;
}

/* This function creates a Hanning window of the requested length,
   allocating the necessary storage, and returns a pointer to that
   storage.  The storage can be freed once the window is no longer
   needed by calling free(). */
double *hanning(long N)
{
  long i;
  double *h = (double *) malloc(N * sizeof(*h));
  if (!h) 
    fatal("Allocation failure");

  for (i = 0; i < N; i++) {
    double arg;
    arg = (double) 2*M_PI*(i+1);
    arg = arg / (N+1);
    h[i] = 0.5 * (1 - cos(arg));
  }

  return h;
}

/* In-place FFT.  Use is==1 for forward FFT, and is==-1 
   for IFFT. */
void call_fft(double *fr, double *fi, unsigned long N, int is)
{ 
  unsigned long log2N;

  // Find the log to base 2 of N.  This assumes
  // that N is a power of 2. 
  log2N = (unsigned long) round(log(N)/log(2));

  fht_fft(fr, fi, log2N, is);
}

/* This writes a sound file. */
void writeSoundFile(double *output, long len, char *name, double Fs)
{
  SOUND *snd;
  
  snd = sndNew(NULL);
  if (snd == NULL) 
    fatal("Allocation failure");
  
  sndSetChans(snd, 1);
  sndSetFormat(snd, SFMT_FLOAT);
  sndSetSrate(snd, Fs);
  sndSetSFformat(snd, SFF_RAW);
  sndSetUformat(snd, SFMT_DOUBLE);
  
  if (sndOpen(snd, name, "w") == NULL) 
    fatal("Unable to open sound file %s.", name);
  
  if (sndWriteFrames(snd, output, len) != len) 
    fatal("Can't write all sample data to output file %s", name);
  
  sndClose(snd);
  sndFree(snd);
}

