#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <QN_args.h>

/* We do audio file I/O using Dan Ellis' snd library. */
extern "C" 
{
#include <snd.h>
#include <sndf.h>
}

#include "meansub-util.h"

/* Analysis window length in seconds. The term "frame" is used in
   other comments to refer to the data inside an analysis window.  The
   window step is fixed at winLength/4, because this makes resynthesis
   easy if we use Hanning windows for analysis--we just need to take
   the IFFT of each frame and add overlaps to resynthesize. */
double winLength = 1.024;

/* Number of past/future frames to use, together with present frame,
   to calculate the mean. Set these both to -1 to calculate the mean 
   over all frames of input data. */
int leftWing = 22;
int rightWing = 22;

int zpad = 0;
int xpad = 0;

char *silflagfile = NULL;

int verbose = 0;

QN_ArgEntry args[] =
  {
    { 0,
      "\nThis program performs log spectral magnitude mean subtraction on input audio"
      "\nand outputs processed audio. To perform just one mean calculation for all audio"
      "\n(instead of a sliding mean-calculation window) use the command line options"
      "\nleft=-1 and right=-1.   To group multiple files together for mean calculation,"
      "\nspecify more than one input file on the command line. The output file(s) will"
      "\nbe placed in the output directory named on the command line.  The output"
      "\ndirectory must be created before this program is run."
      "\n\nOptions:", QN_ARG_DESC},
    { "silflagfile", 
      "Silence flag file (use this to exclude nonspeech frames from the mean calculation).",
      QN_ARG_STR, &silflagfile },
    { "left", 
      "Number of frames of left context to use in mean estimation.",
      QN_ARG_INT, &leftWing },
    { "right", 
      "Number of frames of right context to use in mean estimation.",
      QN_ARG_INT, &rightWing },
    { "win", "Frame length (duration of the spectral analysis window), in seconds.", 
      QN_ARG_DOUBLE, &winLength },
    { "zpad",
      "Make pad frames using zeros, instead of duplicating input samples.", 
      QN_ARG_BOOL, &zpad }, 
    { "xpad",
      "Exclude pad frames from spectral mean calculation (regardless of the value of this option, the pad frames are always excluded if the silflag option is used).", 
      QN_ARG_BOOL, &xpad }, 
    { "verbose",
      "More logging output.", 
      QN_ARG_BOOL, &verbose }, 
    { 0,0,QN_ARG_NOMOREARGS   } 
  };



/* Stores log magnitude spectra.  This is a multi-dimensional array,
   each row of which is the spectrum of a frame.  */
double **logSpectra = NULL;
/* Stores phase spectra. */
double **phaseSpectra = NULL;
/* Store real and imaginary parts of spectra. */
double **spectraRe = NULL;
double **spectraIm = NULL;

/* This function normalizes the log magnitude spectra by mean subtraction.
   The mean used to normalize the spectrum of a particular frame is calculated
   using that frame, leftWing earlier frames, and rightWing future frames. 
   Frames that do not have leftWing frames of left context and rightWing
   frames of right context use the mean calculated for the nearest 
   frame that does.  The user is responsible for making sure at least
   one frame has enough context (this function will abort the program
   otherwise). */
void
normalize(long N, long nFrames, long nPadFrames) 
{
  long i, frame, fr;
  double **logMeans = (double **) malloc((nFrames) * sizeof(*logMeans));


  /* Number of frames at start and end to exclude from mean calculation. 
     This is used to implement the xpad option to exclude pad frames
     from the mean calculation.  */
  long nskip = 0; 
  if (xpad) nskip = nPadFrames;

  /* Allocate storage. */
  if (!logMeans)
    fatal("Allocation failure");

  if (nFrames < leftWing + 1 + rightWing + 2*nskip)
    fatal("Not enough input data to estimate mean.");

  /* Calculate mean spectra for each frame that has enough
     left and right context. */
  for (frame = leftWing + nskip; frame < nFrames - rightWing - nskip; frame++)
    {
      double *mean = (double *) malloc(N * sizeof(*mean));

      if (!mean)
	fatal("Allocation failure");

      for (i = 0; i < N; i++) 
	mean[i] = 0.0;

      for (fr=frame-leftWing; fr<=frame+rightWing; fr++) {
	for (i = 0; i < N; i++) {
	  mean[i] += logSpectra[fr][i];
	}
      }
    
      for (i = 0; i < N; i++) 
	mean[i] = mean[i] / (leftWing+rightWing+1);

      logMeans[frame] = mean;
    }

  /* For frames that do not have enough context for the mean calculation,
     use the mean calculated for the nearest frame that has enough context. */
  for (frame = 0; frame < leftWing + nskip; frame++)
    logMeans[frame] = logMeans[leftWing + nskip];
  for (frame = nFrames-rightWing - nskip; frame < nFrames; frame++)
    logMeans[frame] = logMeans[nFrames-rightWing-nskip-1];

  /* Mean subtraction. */
  for (frame = 0; frame < nFrames; frame++)
    {
      for (i = 0; i < N; i++)  
	{
	  logSpectra[frame][i] -= logMeans[frame][i];
	}
    }
    
    
  /* Free storage. */
  for (frame = leftWing+nskip; frame < nFrames - rightWing-nskip; frame++) 
    free(logMeans[frame]);

  free(logMeans);
}

/* This function normalizes the log magnitude spectra by mean subtraction.
   The mean used to normalize the spectra is calculated only once, using
   all the frames in the utterance. */
void
normalize_entire(long N, long nFrames, long nPadFrames) 
{
  long i, frame;
  double *mean = (double *) malloc(N * sizeof(*mean));

  /* Number of frames at start and end to exclude from mean calculation. 
     This is used to implement the xpad option to exclude pad frames
     from the mean calculation.  */
  long nskip = 0; 
  if (xpad) nskip = nPadFrames;

  if (!mean)
    fatal("Allocation failure");

  for (i = 0; i < N; i++) 
    mean[i] = 0.0;

  for (frame = 0 + nskip; frame < nFrames - nskip; frame++) {
    for (i = 0; i < N; i++) {
      mean[i] += logSpectra[frame][i];
    }
  }
    
  for (i = 0; i < N; i++) 
    mean[i] = mean[i] / nFrames;

  /* Mean subtraction. */
  for (frame = 0; frame < nFrames; frame++)
    {
      for (i = 0; i < N; i++)  
	{
	  logSpectra[frame][i] -= mean[i];
	}
    }  
    
  free(mean);
}

/* This function normalizes the log magnitude spectra by mean subtraction.
   The mean used to normalize the spectra is calculated only once, using
   all the frames in the utterance. */
void
normalize_entire_silflag(long N, long nFrames, char *silflag, long nPadFrames) 
{
  long i, frame;
  double *mean = (double *) malloc(N * sizeof(*mean));

  /* Number of frames at start and end to exclude from mean calculation. 
     This is used to implement the xpad option to exclude pad frames
     from the mean calculation.  */
  long nskip = 0; 
  if (xpad) nskip = nPadFrames;

  if (!mean)
    fatal("Allocation failure");

  for (i = 0; i < N; i++) 
    mean[i] = 0.0;

  int numSpeechFrames = 0;
  for (frame = 0 + nskip; frame < nFrames - nskip; frame++) {
    if (!silflag[frame])
      {
	numSpeechFrames++;
    
	for (i = 0; i < N; i++) {
	  mean[i] += logSpectra[frame][i];
	}
      }
  }

  if (numSpeechFrames==0)
    fatal("numSpeechFrames==0");
  
  for (i = 0; i < N; i++)
    mean[i] = mean[i] / numSpeechFrames;

  /* Mean subtraction. */
  for (frame = 0; frame < nFrames; frame++)
    {
      for (i = 0; i < N; i++)  
	{
	  logSpectra[frame][i] -= mean[i];
	}
    }  
    
  free(mean);
}

void
normalize_silflag(long N, long nFrames, char *silflag, long nPadFrames) 
{
  long i, frame, fr;
  double **logMeans = (double **) malloc((nFrames) * sizeof(*logMeans));

  /* Number of frames at start and end to exclude from mean calculation. 
     This is used to implement the xpad option to exclude pad frames
     from the mean calculation.  */
  long nskip = 0; 
  if (xpad) nskip = nPadFrames;

  /* Allocate storage. */
  if (!logMeans)
    fatal("Allocation failure");

  if (nFrames < leftWing + 1 + rightWing + 2*nskip)
    fatal("Not enough input data to estimate mean.");

  /* Calculate mean spectra for each frame that has enough
     left and right context. */
  for (frame = leftWing + nskip; frame < nFrames - rightWing - nskip; frame++)
    {
      double *mean = (double *) malloc(N * sizeof(*mean));

      if (!mean)
	fatal("Allocation failure");

      for (i = 0; i < N; i++) 
	mean[i] = 0.0;

      int numSpeechFrames = 0;
      for (fr=frame-leftWing; fr<=frame+rightWing; fr++) {
	if (!silflag[fr])
	  {
	    numSpeechFrames++;

	    for (i = 0; i < N; i++) {
	      mean[i] += logSpectra[fr][i];
	    }
	  }
      }
    
      if (numSpeechFrames==0)
	fatal("numSpeechFrames==0");
  
      for (i = 0; i < N; i++) 
	mean[i] = mean[i] / numSpeechFrames;

      logMeans[frame] = mean;
    }

  /* For frames that do not have enough context for the mean calculation,
     use the mean calculated for the nearest frame that has enough context. */
  for (frame = 0; frame < leftWing + nskip; frame++)
    logMeans[frame] = logMeans[leftWing + nskip];
  for (frame = nFrames-rightWing - nskip; frame < nFrames; frame++)
    logMeans[frame] = logMeans[nFrames-rightWing-nskip-1];

  /* Mean subtraction. */
  for (frame = 0; frame < nFrames; frame++)
    {
      for (i = 0; i < N; i++)  
	{
	  logSpectra[frame][i] -= logMeans[frame][i];
	}
    }
    
    
  /* Free storage. */
  for (frame = leftWing+nskip; frame < nFrames - rightWing-nskip; frame++) 
    free(logMeans[frame]);

  free(logMeans);
}

/*
  Read silence flags ('0' or '1' for each frame; '1' = silence) from file
  silfilename.
  
  Allocates and returns a vector of 0s and 1s (not '0's and '1's!).

  Length of the vector is stored to store_n.

  nPadFrames argument is necessary to pad out the vector by
  nPadFrames on each side.  Flags for those frames will be set to 1
  (automatic exclusion from mean calculation).
  
*/  

char *read_silflag(char *silfilename, int *store_n, long nPadFrames)
{
  int c;
  int n = 0; /* number of frames in silence file */
  int f = 0; /* frame counter */
  FILE *silfile;
  char *silflag;

  silfile = fopen(silfilename, "r");

  if (silfile == NULL) 
    fatal("Cannot open silence flag file %s", silfilename);

  while (((c = fgetc(silfile)) != EOF) && (c != '\n')) {
    n++;
  }

  silflag = (char *) malloc( (n) * sizeof(char) + 2*nPadFrames + 1);
  if (silflag == NULL) 
    fatal("Cannot allocate %d bytes", n);

  for (f = 0; f < nPadFrames; f++) {
    silflag[f] = 1; // default to nonspeech
  }

  rewind(silfile);
  for (f = nPadFrames; f < n + nPadFrames; f++) {
    c = fgetc(silfile);
    if ((c != '0') && (c != '1')) 
      fatal("Bad character '%c' (%d) in silence flag file", c, c);
    silflag[f] = (char) (c - '0');
  }
  fclose(silfile);

  for (f = n + nPadFrames; f < n + 2*nPadFrames + 1; f++) {
    silflag[f] = 1; // default to nonspeech
  }

  *store_n=n+2*nPadFrames + 1;
  return silflag;
}


/* This function resynthesizes a time-domain signal from the 
   real and imaginary parts of the frame spectra.  */ 
void
resynth(long N, long nFrames, double *outputWithPad, double *output, long nPadFrames) 
{
  long i, i2, frame, lengthWithPad, startIndex;

  /* code assumes step is N/4 */
  lengthWithPad = (N/4)*(nFrames+3);

  /* Initialize output array to zero. */
  for (i = 0; i < lengthWithPad; i++) { 
    outputWithPad[i] = 0.0; 
  }
  
  /* We take the IFFT of each frame spectrum and add overlaps.  The factor
     of 0.5 is associated with the use of a Hanning window during analysis.  */
  startIndex = 0;   
  for (frame = 0; frame < nFrames; frame++) {
    call_fft(spectraRe[frame], spectraIm[frame], N, -1); 
    
    for (i = 0; i < N; i++) {
      outputWithPad[startIndex + i] += 0.5 * spectraRe[frame][i];
    }  

    startIndex = startIndex + N/4;
  }

  //printf("resynth: lengthWithPad = %ld\n", lengthWithPad);


  /* Throw away the pad frames that we assume process() has added.
     While we're at it, scale by 1/N since that is not performed
     by the FFT routine.  */
  i2 = 0;
  for (i = nPadFrames*N/4; i < lengthWithPad - nPadFrames*N/4; i++) {
    output[i2] = ((double) 1/N)*outputWithPad[i];
    i2++;  
  }  

  //printf("resynth: length = %ld\n", i2);
}

void
process(double *input, long nsamples, long N, long step)
{
  long nSpectra, nFrames, nPadFrames, nalloc, i, startIndex, frame;
  double *h, *input2;

  /* Initialize hanning window. */
  h=hanning(N);

  nSpectra = (long) round(1 + ((double) nsamples - N)/step);
  if (verbose) 
    status("%ld frames.\n", nSpectra);

  /* Idea: Resynthesis is easy with Hanning window using a 4:1 overlap
     ratio, since all that needs to be done is adding 4 IFFT'd windows.
     But, at the edges there aren't 4 overlapped windows.  So we pad
     around the edges.  Then after resynthesis, we throw away the pad
     samples. */

  nPadFrames = 3; 
  nFrames = nSpectra+2*nPadFrames;

  logSpectra = (double **) malloc((nFrames) * sizeof(*logSpectra));  
  phaseSpectra = (double **) malloc((nFrames) * sizeof(*phaseSpectra));

  if (!phaseSpectra || !logSpectra)
    fatal("Allocation failure");

  nalloc = nsamples + 2*step*nPadFrames;

  if ((input2 = (double *) malloc(nalloc * sizeof(*input2))) == NULL)
    fatal("Allocation failure");

  /* leading pad */
  for (i = 0; i < step*nPadFrames; i++) 
    {
      if (zpad)
	input2[i] = 0;
      else
	input2[i] = input[i];
    }

  /* data */
  for (i = 0; i < nsamples; i++)
    input2[step*nPadFrames+i] = input[i];

  /* trailing pad */
  for (i = 0; i < step*nPadFrames; i++)
    {
      if (zpad)
	input2[nsamples+step*nPadFrames+i] = 0;
      else 
	input2[nsamples+step*nPadFrames+i] = input[nsamples - step*nPadFrames + i];
    }
  startIndex = 0;

  for (frame = 0; frame < nFrames; frame++) {
    double re[N], im[N];

    double *lS = (double *) malloc(N * sizeof(*lS)); 
    double *pS = (double *) malloc(N * sizeof(*pS)); 

    if (!lS || !pS) 
      fatal("Allocation failure");

    for (i = 0; i < N; i++) { 
      re[i] = input2[startIndex+i] * h[i];
      im[i] = 0.0;
    }

    call_fft(re, im, N, 1);

    for (i = 0; i < N; i++) {
      if ( re[i]*re[i] + im[i]*im[i] == 0)
	{
	  printf("Using floor to avoid log(0) (frame %ld, bin %ld)\n", frame, i);

          // pick something small
	  lS[i] = -20;
	}
      else
	lS[i] = 0.5*log( re[i]*re[i] + im[i]*im[i] );
      pS[i] = atan2(im[i], re[i]);
    }

    logSpectra[frame] = lS;
    phaseSpectra[frame] = pS;

    startIndex = startIndex + step;
  }

  free(h);

  if (silflagfile == NULL)
    {
      if ((leftWing == -1) && (rightWing == -1))
	{
	  /* Using entire utterance for mean calculation. */
	  normalize_entire(N, nFrames, nPadFrames);
	}	  
      else
	{
	  // default case
	  normalize(N, nFrames, nPadFrames);
	}
    }
  else
    {
      int nFlags;
      char *silflags;
      silflags = read_silflag(silflagfile, &nFlags, nPadFrames);
	  
      /* If we use noise reduced input data to this program, it may be
	 that nFlags is one or two greater than nFrames because the
	 noise reduction program may drop some data to get to an integer
	 number of frames. */
      if (nFlags != nFrames)
	{
	  if (nFlags == nFrames + 1)
	    {
	      // OK
	    }
	  else if (nFlags == nFrames + 2)
	    {
	      // OK
	    }
	  else
	    fatal("%d flags in the flag file but %d audio frames", 
		  nFlags, nFrames);
	}

      if ((leftWing == -1) && (rightWing == -1))
	{
	  normalize_entire_silflag(N, nFrames, silflags, nPadFrames);
	
	}	  
      else
	{
	  normalize_silflag(N, nFrames, silflags, nPadFrames);
	}
    }

  spectraRe = (double **) malloc(nFrames * sizeof(*spectraRe));
  spectraIm = (double **) malloc(nFrames * sizeof(*spectraIm));
  if (!spectraRe || !spectraIm)
    fatal("Allocation failure");

  for (frame = 0; frame < nFrames; frame++) {
    spectraRe[frame] = (double *) malloc(N * sizeof(**spectraRe)); 
    spectraIm[frame] = (double *) malloc(N * sizeof(**spectraIm)); 
    if (!spectraRe[frame] || !spectraIm[frame])
      fatal("Allocation failure");

    for (i = 0; i < N; i++) { 
      double mag;

      mag = exp(logSpectra[frame][i]);
      
      spectraRe[frame][i] = mag * cos(phaseSpectra[frame][i]);
      spectraIm[frame][i] = mag * sin(phaseSpectra[frame][i]);      

    }    
  }  


  resynth(N, nFrames, input2, input, nPadFrames);
  
  free(input2);

  for (frame = 0; frame < nFrames; frame++) {
    free(logSpectra[frame]);
    free(phaseSpectra[frame]);
    free(spectraRe[frame]);
    free(spectraIm[frame]);
  }
  free(logSpectra);
  free(phaseSpectra);
  free(spectraRe);
  free(spectraIm);
}

int 
main(int argc, const char **argv) {

  const char  *sOutDir;
  SOUND *snd;
  double Fs = 0;
  long N, step, npad, nalloc, nsamples, i;
  double *input = 0;
  double *output = 0;
  int numInputFiles; 
  long *sizes;
  char **names;
  int SFformat = 0;
  char *appname;

  QN_initargs(args,&argc, &argv, &appname);

  if (argc < 2) 
    fatal("Usage: %s [options] <file1> <file2> ...  <output directory>\n"
	  "Run '%s help=y' for options documentation.", 
	  appname, appname);

  numInputFiles = argc - 1;
  
  sizes = (long *) malloc( numInputFiles * sizeof(*sizes) );
  names = (char **) malloc( numInputFiles * sizeof(*names) );

  if (!sizes || !names)
    fatal("Allocation failure");

  nsamples = 0;

  for (i = 0; i < numInputFiles ; i++) {
    char *p;
    double *readp;

    snd = sndOpen(NULL, (char *) argv[i], "r");
    if (snd == NULL) 
      fatal("Unable to open sound file %s.", argv[i]);

    // Tell the snd library that we want to treat the
    // audio data as double-precision floating point.
    sndSetUformat(snd, SFMT_DOUBLE);

    // If this is the first file
    if (i == 0) {
      SFformat = sndGetSFformat(snd);
      
      if (SFformat == SFF_RAW){
	char *s = getenv("PCMFORMAT");
	
	if (!s) 
	  fatal("If the input files are headerless, you must use the PCMFORMAT environment variable.\n");
      }

      Fs = sndGetSrate(snd);
      status("Sampling rate = %g Hz.\n", Fs);
    }
    // Else, this is not the first file
    else if ((SFformat != sndGetSFformat(snd)) ||
	     (Fs != sndGetSrate(snd))) {
      /* It seems that if I open a series of files in succession with sndOpen,
	 sndGetSrate will keep returning the sampling rate of the first file
	 even if latter files have different sampling rates.  So this check
	 does not work reliably. */
      fatal("All input files must have the same sample format and sampling rate.");
    }

    sizes[i] = sndGetFrames(snd);
    p = strrchr(argv[i], '/');
    if (p == NULL)
      names[i] = strdup(argv[i]);
    else
      names[i] = strdup(p+1);
    if (!names[i])
      fatal("Allocation failure");    
    if (verbose)
      status("Opened file %s: %ld samples.\n", names[i], sizes[i]); 

    nsamples += sizes[i];

    if (i == 0) {   
      if ((input = (double *) malloc(nsamples * sizeof(*input))) == NULL)
	fatal("Allocation failure");
    } else {
      if ((input =
	   (double *) realloc(input, nsamples * sizeof(*input))) == NULL)
	fatal("Allocation failure"); 
    }

    readp = &input[nsamples - sizes[i]];

    if (sndReadFrames(snd, readp, sizes[i]) != sizes[i]) 
      fatal("Can't read all of input file %s", argv[i]);
   
    sndFree(snd);
  }

  status("Opened %d input file(s)\n", numInputFiles);

  N = (long)round(Fs*winLength);

  step = N/4;

  status("Frame length = %f s (%ld samples), frame step = %f s (%ld samples)\n", 
	 winLength, N, winLength/4, N/4);

  /* Check if N is a power of 2.  Casting N to unsigned is safe here,
   since unsigned and signed representations are identical for the
   range of values we expect for N. */
  if (!isPowerOf2( (unsigned long) N)) {
    fatal("The frame length is not a power of 2.\n"
	  "This could cause problems with the fht_fft routine we use for the FFT.");
  }
  
  if ((leftWing == -1) && (rightWing == -1)) {
    status("Using entire utterance for mean calculation\n");
  } else {
    status("Using %f s of data for mean calculation (left context %d frames, right context %d frames)\n",
	   winLength + (leftWing+rightWing)*winLength/4, leftWing, rightWing);
  }

  /* Make sure there are an integer number of window steps in the
     sample by padding at the end if necessary.  (This is separate from
     the padding performed by the process() routine.) */
  if (nsamples < N)
    npad = (long) (N - nsamples);
  else 
    {
      npad = (long) (step - fmod( nsamples - N, step ));
      if (npad == step)
	npad = 0; /* no padding needed */
    }
  if (verbose)
    status("%ld input samples total, duplicating %ld samples at tail to make ", 
	   nsamples, npad);
  
  nalloc = nsamples + npad;

  if ((input =
       (double *) realloc(input, nalloc * sizeof(*input))) == NULL)
    fatal("Allocation failure"); 

  if (npad != 0) {

    /* Duplicate input samples as pad samples so that we have an
       integer number of analysis frames.  We use modulo arithmetic so
       that if we need more pad samples than there are input samples
       we wrap around the input sample buffer.  */

    for (i = 0; i < npad; i++) {
      int x, y, z;

      x = nsamples - npad + i;
      y = nsamples;

      z = x % y;
      if (z < 0) z += y;   

      input[nsamples+i] = input[z]; 
    }
  }

  process(input, nalloc, N, step);

  output = input; 

  /* output directory */
  sOutDir = argv[argc-1];
  if (verbose) {
    status("Creating output in %s.\n", sOutDir);
  }
  if (chdir(sOutDir))
    fatal("Can't cd to output directory.");

  /* The following two lines are to deal with an bug (apparently) in
     the sndf library where if the input files are NIST format,
     and we try to write out raw floating point files, the first
     file written out will actually have a NIST header.   This
     code was written in 2001 so the sndf library may not
     still have this problem. */
  i = 0;
  writeSoundFile(output, sizes[i], names[i], Fs);  
  
  for (i = 0; i < numInputFiles; i++) 
    {
      writeSoundFile(output, sizes[i], names[i], Fs);
      output += sizes[i];
    }


  free(input);
  free(sizes);
  for (i = 0; i < numInputFiles; i++)
    free(names[i]);
  free(names);

  return 0;
}





